"""
Functions for running the Mast3r inference pipeline.
"""

__all__ = (
    #"symmetric_inference",
    #"pairs_inference",
    "reconstruct_scene",
)

import tempfile

import numpy as np
import torch
from tqdm import tqdm

from dust3r.image_pairs import make_pairs
from mast3r.cloud_opt.sparse_ga import sparse_global_alignment, extract_correspondences

from .image import prepare_images_for_mast3r
from .scene import PointCloudScene


def symmetric_inference(model, img1, img2) -> dict[str, list[torch.Tensor]]:
    """
    Run inference on a pair of images w.r.t. both images.

    Returns x11, x21, x22, x12.

    x11, x21 are obtained w.r.t. img1.
    - x11 is img1 pointmap in img1 coords; x21 is img2 pointmap in img1 coords.
    - x22, x12 w.r.t. img2, similarly.

    Put model and images on the same torch device externally.

    :param img1, img2: Tensor, shape (3, H, W).
    :return: ``{"pts3d": [...,], "conf": [...,], "desc": [...,], "desc_conf": [...,]}``
        i.e. flattens keys of model output.
        Each value is a list of 4, corresponding to x11, x21, x22, x12.
    """
    img1 = img1.unsqueeze(0)
    img2 = img2.unsqueeze(0)
    shape1 = torch.tensor(img1.shape[-2:]).unsqueeze(0)
    shape2 = torch.tensor(img2.shape[-2:]).unsqueeze(0)

    # compute encoder only once
    feat1, feat2, pos1, pos2 = model._encode_image_pairs(img1, img2, shape1, shape2)

    def decoder(feat1, feat2, pos1, pos2, shape1, shape2):
        dec1, dec2 = model._decoder(feat1, pos1, feat2, pos2)
        with torch.cuda.amp.autocast(enabled=False):
            res1 = model._downstream_head(1, [tok.float() for tok in dec1], shape1)
            res2 = model._downstream_head(2, [tok.float() for tok in dec2], shape2)
        return res1, res2

    # decoder 1-2
    res11, res21 = decoder(feat1, feat2, pos1, pos2, shape1, shape2)
    # decoder 2-1
    res22, res12 = decoder(feat2, feat1, pos2, pos1, shape2, shape1)

    # Flatten keys
    results = [res11, res21, res22, res12]
    keys = ("pts3d", "conf", "desc", "desc_conf")
    ret = {}
    for key in keys:
        ret[key] = [r[key][0] for r in results]
    return ret


@torch.no_grad()
def pairs_inference(model, imgs, pair_indices, verbose=False):
    """
    Run symmetric inference on all pairs of images.

    Returns key-value map of pair indices to results.

    :return:
    {
        (i1, i2): {
            "pts3d": ...,
            "conf": ...,
            "matching_score": ...,
            "corres": ...,
        },
    }
    See original function (mast3r/cloud_opt/sparse_ga.py:forward_mast3r) for details.
    """
    ret = {}

    for i1, i2 in tqdm(pair_indices, disable=not verbose, desc="Pairs inference"):
        output = symmetric_inference(model, imgs[i1], imgs[i2])
        xs = output["pts3d"]
        confs = output["conf"]
        descs = output["desc"]
        qonfs = output["desc_conf"]

        corres = extract_correspondences(descs, qonfs, device=imgs[0].device, subsample=8)
        # Geometric mean of confidences
        conf_score = np.prod([c.mean().item() for c in confs]) ** (1 / 4)
        matching_score = (float(conf_score), float(corres[2].sum()), len(corres[2]))

        ret[(i1, i2)] = {
            "pts3d": xs,
            "conf": confs,
            "matching_score": matching_score,
            "corres": corres,
        }

    return ret


def reconstruct_scene(model, imgs, filelist, device, tmpdir=None) -> PointCloudScene:
    """Run Mast3r reconstruction pipeline.

    Mast3r model inference and global adjustment.

    Parameters
    ----------

    model:
        Model instance :class:`starster.Mast3rModel`.

    imgs:
        List of images from :func:`starster.load_image`.
        Tensor shape (C,H,W), dtype float32.

    filelist:
        List of image paths corresponding to each image.
        Due to Mast3r legacy code, this is required.

    device:
        Torch device to run on.

    tmpdir:
        Temp directory. If None, a new one is created tempfile.

    Returns
    -------

    Reconstructed scene as :class:`PointCloudScene`.
    """
    imgs = prepare_images_for_mast3r(imgs)
    pairs = make_pairs(imgs, scene_graph="complete", prefilter=None, symmetrize=True)

    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    scene = sparse_global_alignment(
        filelist,
        pairs,
        tmpdir,
        model,
        lr1=0.07,
        niter1=500,
        lr2=0.014,
        niter2=200,
        device=device,
        opt_depth=False,
        matching_conf_thr=5,
        shared_intrinsics=False,
    )

    scene = PointCloudScene(scene)

    return scene



"""
Reconstructed point cloud wrapper class.
"""

__all__ = (
    "PointCloudScene",
)

import torch


class PointCloudScene:
    """Reconstructed point cloud; output of Mast3r pipeline.

    Use :func:`starster.reconstruct_scene` to get an instance of this class.

    This class has some wrapper functions around Mast3r SparseGA,
    as well as novel functions.
    """

    num_cams: int

    def __init__(self, sparse_ga):
        """
        Parameters
        ----------

        sparse_ga:
            SparseGA instance from Mast3r code.
        """
        self.sparse_ga = sparse_ga
        self.num_cams = len(self.sparse_ga.pts3d)

    @property
    def imgs(self) -> list[torch.Tensor]:
        """Get source image tensors.

        Alias of ``self.sparse_ga.imgs``.

        Returns
        -------

        List of tensors for each camera.

        Each tensor is shape (H, W, 3).
        """
        return self.sparse_ga.imgs

    def pts_sparse(self) -> list[tuple[torch.Tensor, torch.Tensor]]:
        """Get reconstructed 3D point coordinates and colors.

        These four functions (pts_sparse, pts_sparse_flat, pts_dense, pts_dense_flat)
        behave similarly.

        Sparse and dense return their respective point clouds.

        Flat returns a concatenated point cloud from all cameras together.

        Non-flat returns a list of point clouds, one for each camera.

        Each point cloud is returned as ``(pts3d, colors)``:

        - pts3d: 3D point coordinates. Shape (N, 3).
        - colors: RGB colors for each point. Shape (N, 3).
        """
        ret = []
        for i in range(self.num_cams):
            ret.append((self.sparse_ga.pts3d[i], self.sparse_ga.pts3d_colors[i]))
        return ret

    def pts_sparse_flat(self) -> tuple[torch.Tensor, torch.Tensor]:
        """See ``pts_sparse``."""
        pts = self.pts_sparse()
        pts3d = torch.cat([p[0] for p in pts], dim=0)
        colors = torch.cat([p[1] for p in pts], dim=0)
        return pts3d, colors

    def pts_dense(self, conf_thres=1.5) -> list[tuple[torch.Tensor, torch.Tensor]]:
        """See ``pts_sparse``."""
        ret = []
        pts, _, confs = self.sparse_ga.get_dense_pts3d(clean_depth=True)
        for i in range(self.num_cams):
            colors = torch.tensor(self.sparse_ga.imgs[i]).reshape(-1, 3)
            mask = (confs[i] > conf_thres).reshape(-1).cpu()
            ret.append((pts[i][mask], colors[mask]))
        return ret

    def pts_dense_flat(self, conf_thres=1.5) -> tuple[torch.Tensor, torch.Tensor]:
        """See ``pts_sparse``."""
        pts = self.pts_dense(conf_thres)
        pts3d = torch.cat([p[0] for p in pts], dim=0)
        colors = torch.cat([p[1] for p in pts], dim=0)
        return pts3d, colors

    def c2w(self) -> torch.Tensor:
        """Returns camera-to-world transformation matrices, shape (C, 4, 4).

        Alias of ``self.sparse_ga.cam2w``.
        """
        return self.sparse_ga.cam2w

    def w2c(self) -> torch.Tensor:
        """Returns world-to-camera transformation matrix (inverse of ``c2w``)."""
        return torch.inverse(self.c2w())

    def intrinsics(self) -> torch.Tensor:
        """Returns camera intrinsic matrices, shape (C, 3, 3).

        Alias of ``self.sparse_ga.intrinsics``.
        """
        return self.sparse_ga.intrinsics



"""
Image loading and processing.
"""

__all__ = (
    #"make_pair_indices",
    "process_image",
    "load_image",
    "load_images",
    "prepare_images_for_mast3r",
)

from pathlib import Path

import numpy as np
import torch
import torchvision.transforms as T
from PIL import Image
from PIL.ImageOps import exif_transpose

_to_tensor = T.ToTensor()
_normalize = T.Normalize(mean=(0.5, 0.5, 0.5), std=(0.5, 0.5, 0.5))


def make_pair_indices(n: int, symmetric: bool = True) -> list[tuple[int, int]]:
    """
    Generate all pairs of indices for ``n`` elements.

    Fully connected graph; i.e. (i, j) exists for i, j < n and:
    - Symmetric: i != j;
    - Not symmetric: i < j.
    """
    pairs = []
    for i in range(n):
        for j in range(i):
            pairs.append((i, j))
    if symmetric:
        for i in range(len(pairs)):
            pairs.append((pairs[i][1], pairs[i][0]))
    return pairs


def process_image(img: np.ndarray | torch.Tensor, size: int) -> torch.Tensor:
    """Preprocess image to pipeline requirements.

    - Resizes longest edge of image to ``size``.
    - Crops (around center) H and W to a multiple of 8.

    Parameters
    ----------

    img:
        Shape (C,H,W), dtype uint8

    Returns
    -------

    Tensor, shape (C,H,W), dtype float32
    """
    new_size = [int(x * size / max(img.shape[1:])) for x in img.shape[1:]]
    img = T.functional.resize(img, new_size, T.InterpolationMode.BICUBIC)

    cx = img.shape[2] // 2
    cy = img.shape[1] // 2

    # Compute width and height as mult of 8
    width_half = (cx // 8) * 8
    height_half = (cy // 8) * 8

    # Crop around center
    img = img[..., cy - height_half : cy + height_half, cx - width_half : cx + width_half]

    # Torchvision transforms
    img = _normalize(img)

    return img


def load_image(path: str | Path, size: int = 224) -> torch.Tensor:
    """Load and process image from file.

    Parameters
    ----------

    path:
        Path to image.

    size:
        Resize longest edge to this.

    Returns
    -------

    Tensor. See :func:`process_image`.
    """
    img = Image.open(path)
    img = exif_transpose(img)
    img = img.convert("RGB")
    img = _to_tensor(np.array(img))
    img = process_image(img, size)
    return img


def load_images(paths: list[str | Path], size: int = 224) -> list[torch.Tensor]:
    """Load a list of files.

    Calls :func:`load_image` on each path.
    """
    return [load_image(p, size) for p in paths]


def prepare_images_for_mast3r(imgs: list[torch.Tensor]):
    """
    Returns image format as supported by mast3r legacy code.

    Each image is

    {
        img: Tensor image, shape (1, 3, H, W)
        true_shape: [[H, W]]
        idx: int
        instance: str(idx)
    }

    TODO legacy code should be replaced and this function should be removed.
    """
    ret = []
    for i in range(len(imgs)):
        img = imgs[i]
        ret.append(
            dict(
                img=img[None],
                true_shape=np.int32([img.shape[-2:]]),
                idx=i,
                instance=str(i),
            )
        )

    return ret



"""
3D Gaussian Splatting scene refinement.
"""

__all__ = (
    "GSTrainer",
)

import gsplat
import torch
from torchmetrics.image import StructuralSimilarityIndexMeasure as SSIM
from tqdm import trange

from .scene import PointCloudScene


class GSTrainer:
    """3DGS training and refinement.

    Uses gsplat for rendering and optimization.

    Uses the Mast3r reconstruction as a starting point.
    """

    def __init__(
            self,
            scene: PointCloudScene,
            init_scale=3e-3,
            lr=1e-3,
            loss_ssim_fac=0.2,
            loss_opacity_fac=0.01,
            loss_scale_fac=0.01,
            device: str = "cuda"
        ):
        """Initialize the splats and optimizers from the Mast3r reconstruction.

        Parameters
        ----------

        scene:
            Reconstructed scene from Mast3r.
            Return value of :func:`starster.reconstruct_scene`.

        init_scale:
            Initial scale of the splats.

        lr:
            Learning rate for Adam optimizers.

        loss_ssim_fac:
            Factor for the SSIM loss.

        loss_opacity_fac:
            Factor for the opacity loss.

        loss_scale_fac:
            Factor for the scale loss.

        device:
            Device to use.
        """
        self.scene = scene
        self.device = device

        # Create splats
        pts, colors = self.scene.pts_dense_flat()
        self.gaussians = {
            "means": pts,
            "scales": torch.full_like(pts, init_scale),
            "quats": torch.zeros(pts.shape[0], 4),
            "opacities": torch.ones(pts.shape[0]),
            "sh0": torch.zeros(pts.shape[0], 1, 3),
            "shN": torch.zeros(pts.shape[0], 24, 3),
        }
        self.gaussians["quats"][:, 0] = 1.0
        self.gaussians["sh0"][:, 0] = 1 - colors
        for i in range(24):
            self.gaussians["shN"][:, i] = 1 - colors

        for k, v in self.gaussians.items():
            self.gaussians[k] = torch.nn.Parameter(v.to(self.device))

        # Create optimizers
        self.optimizers = {k: torch.optim.Adam([v], lr=lr) for k, v in self.gaussians.items()}

        # Create pruning strategy
        # Using MC strategy bc DefaultStrategy has a bug.
        self.strategy = gsplat.MCMCStrategy()
        self.strategy.check_sanity(self.gaussians, self.optimizers)
        self.strategy_state = self.strategy.initialize_state()

        # Create required losses
        self.loss_ssim_fac = loss_ssim_fac
        self.loss_opacity_fac = loss_opacity_fac
        self.loss_scale_fac = loss_scale_fac

        self.ssim = SSIM(data_range=1).to(self.device)

    def render_views(self, w2c: torch.Tensor, intrinsics: torch.Tensor, width: int, height: int):
        """Render the splats from a set of camera views.

        Parameters
        ----------

        w2c:
            World-to-camera matrices. Shape (N, 4, 4).

        intrinsics:
            Camera intrinsics. Shape (N, 3, 3).

        width:
            Image width.

        height:
            Image height.

        Returns
        -------

        Output of ``gsplat.rasterization``.

        Tuple of ``(render_img, render_alpha, info)``.

        - render_img: Color image. Shape (N, H, W, 3).
        """
        render = gsplat.rasterization(
            means=self.gaussians["means"],
            quats=self.gaussians["quats"],
            scales=self.gaussians["scales"],
            opacities=self.gaussians["opacities"],
            colors=self.gaussians["shN"],
            viewmats=w2c,
            Ks=intrinsics,
            width=width,
            height=height,
            sh_degree=1,
        )
        return render

    def render_views_original(self, width: int, height: int):
        """Render from camera views of original scene (``self.scene``).

        See ``render_views``.
        """
        return self.render_views(self.scene.w2c(), self.scene.intrinsics(), width, height)

    def compute_loss(self, truth_img, render_img, render_alpha):
        l1 = torch.nn.functional.l1_loss(truth_img, render_img)

        ssim = 1 - self.ssim(truth_img.permute(2, 0, 1).unsqueeze(0), render_img.permute(2, 0, 1).unsqueeze(0))
        loss = l1 * (1 - self.loss_ssim_fac) + ssim * self.loss_ssim_fac

        loss += self.loss_opacity_fac * torch.abs(torch.sigmoid(self.gaussians["opacities"])).mean()

        loss += self.loss_scale_fac * torch.abs(torch.exp(self.gaussians["scales"])).mean()

        return loss

    def run_optimization(self, iters: int, enable_pruning: bool = False, verbose: bool = False) -> list[float]:
        """Run 3DGS optimization and pruning (optional) for a number of iterations.

        Parameters
        ----------

        iters:
            Number of iterations.

        enable_pruning:
            Enable pruning and densification via the gsplat pruning strategy.

        verbose:
            Enable tqdm progress bar.

        Returns
        -------

        List of losses at each iteration.
        """
        height, width = self.scene.imgs[0].shape[:2]

        losses = []

        pbar = trange(iters, disable=not verbose)
        for step in pbar:
            render_img, render_alpha, info = self.render_views_original(width, height)

            if enable_pruning:
                self.strategy.step_pre_backward(self.gaussians, self.optimizers, self.strategy_state, step, info)

            loss = 0
            for i in range(self.scene.num_cams):
                img = torch.tensor(self.scene.imgs[i], device=self.device)
                loss += self.compute_loss(img, render_img[i], render_alpha[i])
            loss.backward()

            desc = f"Gsplat optimization: loss={loss.item()}"
            pbar.set_description(desc)
            losses.append(loss.item())

            for optim in self.optimizers.values():
                optim.step()
                optim.zero_grad(set_to_none=True)

            if enable_pruning:
                self.strategy.step_post_backward(self.gaussians, self.optimizers, self.strategy_state, step, info, 1e-3)

        return losses



"""
Do not run this in the root directory.
Run build.sh

This script will be copied and run from the ``build`` directory,
so all paths are relative to that directory.
"""

import setuptools


def read_reqs(path):
    with open(path, "r") as fp:
        return fp.read().splitlines()


with open("../README.md", "r") as fp:
    long_description = fp.read()

requirements = []
req_paths = (
    "../requirements.txt",
    "../mast3r/requirements.txt",
    "../mast3r/dust3r/requirements.txt",
)
for path in req_paths:
    requirements.extend(read_reqs(path))

setuptools.setup(
    name="starst3r",
    version="0.3.0",
    author="Patrick Huang",
    author_email="phuang1024@gmail.com",
    description="Ultra fast 3D reconstruction and novel view synthesis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phuang1024/Starst3r",
    py_modules=["starster", "mast3r", "dust3r", "croco"],
    packages=setuptools.find_packages(),
    install_requires=requirements,
)
